#!/usr/bin/env python

###############################################################################
# Copyright (c) 2010, Vadim Shlyakhov
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#******************************************************************************

import os
import logging
import itertools
from optparse import OptionParser
from subprocess import Popen, PIPE

datum_map={ 
    # http://www.hemanavigator.com.au/Products/TopographicalGPS/OziExplorer/tabid/69/Default.aspx
    'WGS 84':
        '+datum=WGS84',
    'Pulkovo 1942 (1)': # http://trac.osgeo.org/gdal/ticket/3176
        '+ellps=krass +towgs84=23.9,-141.3,-80.9,0,-0.37,-0.85,-0.12', 
    'Pulkovo 1942 (2)':
        '+ellps=krass +towgs84=23.9,-141.3,-80.9,0,-0.37,-0.85,-0.12', 
    'Pulkovo 1942': # http://ne-grusti.narod.ru/oziexplorer.html
        '+ellps=krass +towgs84=23.9,-141.3,-80.9,0,-0.37,-0.85,-0.12', 
    'Australian Geodetic 1966': # http://www.osgeo.org/pipermail/gdal-dev/2006-February/008090.html
        '+ellps=aust_SA +towgs84=-129.193,-41.212,130.73,-0.246,-0.374,-0.329,-2.955',
    'Australian Geodetic 1984':
        '+ellps=aust_SA +towgs84=-117.763,-51.51,139.061,-0.292,-0.443,-0.277,-0.191',
    'Australian Geocentric 1994 (GDA94)':
        '+ellps=GRS80 +towgs84=0,0,0,0,0,0,0',
    }
proj_map={
    'Latitude/Longitude':                       '+proj=latlong',
    'Mercator':                                 '+proj=merc',
    'Transverse Mercator':                      '+proj=tmerc',
    '(UTM) Universal Transverse Mercator':      '+proj=utm',
    '(BNG) British National Grid':              '+init=epsg:27700', # not tested
    '(IG) Irish Grid':                          '+init=epsg:29902', # not tested
    '(NZG) New Zealand Grid':                   '+init=epsg:27200', # not tested
    '(SG) Swedish Grid':                        '+init=epsg:3847',  # not tested
    '(SUI) Swiss Grid':                         '+init=epsg:21781', # not tested
    '(I) France Zone I':                        '+init=epsg:27571', # not tested
    '(II) France Zone II':                      '+init=epsg:27572', # not tested
    '(III) France Zone III':                    '+init=epsg:27573', # not tested
    '(IV) France Zone IV':                      '+init=epsg:27574', # not tested
    'Lambert Conformal Conic':                  '+proj=lcc',        # not tested
    '(A)Lambert Azimuthual Equal Area':         '+proj=laea',       # not tested
    '(EQC) Equidistant Conic':                  '+proj=eqdc',
    'Sinusoidal':                               '+proj=sinu',       # not tested
    'Polyconic (American)':                     '+proj=poly',       # not tested
    'Albers Equal Area':                        '+proj=aea',        # not tested
    'Van Der Grinten':                          '+proj=vandg',      # not tested
    'Vertical Near-Sided Perspective':          '+proj=nsper',      # not tested
    '(WIV) Wagner IV':                          '+proj=wag4',       # not tested
    'Bonne':                                    '+proj=bonne',      # not tested
    '(MT0) Montana State Plane Zone 2500':      '+init=esri:102300',# not tested
    '(ITA1) Italy Grid Zone 1': # http://mpa.itc.it/markus/shortcourse/notes2.html
        '+init=epsg:26591 +towgs84=-85.88,-28.85,+49.67,-1.003,-2.383,-1.808,-27.82', # not tested
    '(ITA2) Italy Grid Zone 2':
        '+init=epsg:26592 +towgs84=-85.88,-28.85,+49.67,-1.003,-2.383,-1.808,-27.82', # not tested
    '(VICMAP-TM) Victoria Aust.(pseudo AMG)': # http://www.gpsoz.com.au/VicRoadsInfo.htm
        '+proj=tmerc +lat_0=145 +x_0=500000 +y_0=10000000',         # not tested
    '(VICGRID) Victoria Australia':             '+init=epsg:3110',  # not tested
    '(VG94) VICGRID94 Victoria Australia':      '+init=epsg:3111',  # not tested
    }
parm_map=( 
    '+lat_0=', # 1. Latitude Origin
    '+lon_0=', # 2. Longitude Origin
    '+k=',     # 3. K Factor
    '+x_0=',   # 4. False Easting
    '+y_0=',   # 5. False Northing
    '+lat_1=', # 6. Latitude 1
    '+lat_2=', # 7. Latitude 2
    '+h=',     # 8. Height - used in the Vertical Near-Sided Perspective Projection
               # 9. Sat - not used
               #10. Path - not used
    )

def l_d(smth):
    logging.debug(str(smth))
    
def command(params,stdin=None):
    l_d(' '.join(params))
    process=Popen(params,stdin=PIPE, stdout=PIPE, stderr=PIPE, universal_newlines=True)
    out=process.communicate(stdin)
    if out[1]: l_d(out[1])
    if process.returncode != 0: raise Exception("*** External program failed: %s" % params[0])
    return out[0]

def flatten(listOfLists): 
    return list(itertools.chain.from_iterable(listOfLists))

def dms2dec(degs,mins,ne):
    return repr((eval(degs)+eval(mins)/60)*(-1 if ne in ('W','S') else 1 ))

def hdr_parms(header, patt): 
    'filter header for params starting with patt '
    return [item for item in header if item[0].startswith(patt)]
    
def hdr_read(map_path): 
    'read map header'
    f=open(map_path, 'r')
    header=[[i.strip() for i in l.split(',')] for l in f]
    del f
    if not (header and header[0][0].startswith('OziExplorer Map Data File')): 
        raise Exception("*** invalid file: %s" % map_path)
    l_d(header)
    return header

def get_srs(hdr, refs):
    'returns srs for the map projection and srs for the REF points'
    datum_id=hdr[4][0]
    proj_id=hdr_parms(hdr, 'Map Projection')[0][1]
    parm_lst=hdr_parms(hdr, 'Projection Setup')[0]
    try:
        datum=datum_map[datum_id].split()
        srs=proj_map[proj_id].split()
    except KeyError: raise Exception("*** Unsupported datum or projection")
    if srs[0].startswith('+proj='): # overwise assume it already has a full data defined
        # get projection parameters
        parms=[ i[0]+i[1] for i in zip(parm_map,parm_lst[1:]) if not i[1] == '']
        if srs[0] == '+proj=utm':
            if refs[0][13] != '': # refs are cartesian with a zone defined
                parms.append('+zone='+refs[0][13])
                if refs[0][16] != 'N': 
                    parms.append('+south')
            else: # refs are lat/long, then find zone, hemisphere
                zone=(eval(dms2dec(*refs[0][9:12]) + 3) % 360) // 6
                parms.append('+zone=%d' % zone)
                if refs[0][8] != 'N': 
                    parms.append('+south')
        srs = srs + parms + datum + ['+no_defs']
    l_d(srs)
    return srs

def find_image(img_path, map_dir):
    imp_path_slashed=img_path.replace('\\','/') # get rid of windows separators
    imp_path_lst=imp_path_slashed.split('/')
    img_patt=imp_path_lst[-1].lower()
    match=[i for i in os.listdir(map_dir if map_dir else '.') if i.lower() == img_patt]
    try:
        return os.path.join(map_dir, match[0])
    except IndexError: raise Exception("*** Image file not found: %s" % img_path)

def map2vrt(map_path):
    (map_dir, map_fname)=os.path.split(map_path)
    base=os.path.join(options.dest_dir, os.path.splitext(map_fname)[0])
    aux_vrt= base+'.aux.vrt'    # auxilary VRT file
    out_vrt= base+'.vrt'        # output VRT file
    if os.path.exists(out_vrt): os.remove(out_vrt)

    hdr=hdr_read(map_path)      # Read map header
    img_file=find_image(hdr[2][0],map_dir)
    l_d(img_file)

    refs=[i for i in hdr_parms(hdr, 'Point') if i[2] != ''] # Get a list of geo refs
    out_srs=get_srs(hdr, refs)
    if refs[0][14] != '': # refs are cartesian
        ref_coord=[i[14:16]for i in refs]
    else: # refs are lat/long
        ref_coord=[(dms2dec(*i[9:12]), dms2dec(*i[6:9])) for i in refs]
        if not out_srs[0].startswith('+proj=latlong'): # reproject coordinates
            lat_long = '\n'.join([i[0]+' '+i[1] for i in ref_coord])
            ref_coord= [ i.split() for i in 
                command(['proj'] + out_srs, lat_long ).splitlines()]
    if len(refs) == 2: # only 2 gcps; assume the chart is north alligned
        refs.append(['','',refs[0][2],refs[1][3]])
        ref_coord.append([ref_coord[0][0],ref_coord[1][1]])
    l_d([refs,ref_coord])
    gcps=flatten([('-gcp', i[0][2],i[0][3],i[1][0],i[1][1]) for i in zip(refs, ref_coord)])
    
    transl_cmd=['gdal_translate','-of','VRT',img_file, aux_vrt,'-a_srs', ' '.join(out_srs)]
    gdalinfo=command(['gdalinfo',img_file])
    if 'Color Table' in gdalinfo:
        transl_cmd=transl_cmd+['-expand','rgb']
    transl_out=command(transl_cmd + gcps)
    print transl_out

    (width, height)=[i.strip() for i in transl_out[len('Input file size is'):].split(',')] # raster width, height
    mmpxy=hdr_parms(hdr, 'MMPXY') # Moving Map border pixels
    mmpll=hdr_parms(hdr, 'MMPLL') # Moving Map border lat,lon
    inside=[i for i in mmpxy # check if MMP polygon is inside the image border
        if not ((i[2] == '0' or i[2] == width) and (i[3] == '0' or i[3] == height))]
    warp_opt=[]
    if options.cut and inside: # polygon is there, then cut
        poly='POLYGON(('+','.join(['%s %s' % (i[2],i[3]) for i in mmpxy])+'))' # Create cutline
        warp_opt.extend(['-wo', 'CUTLINE='+poly])
    if options.no_data:
        nodata=' '.join(options.no_data.split(','))
        warp_opt.extend(['-srcnodata', nodata, '-wo', 'UNIFIED_SRC_NODATA=YES'])
    if warp_opt != []:
        command(['gdalwarp','-wo', 'INIT_DEST=NO_DATA','-dstalpha','-of','VRT', aux_vrt, out_vrt]+warp_opt)
    else:
        os.rename(aux_vrt, out_vrt)

if __name__=='__main__':
    usage = "usage: %prog [--cut] [--dest-dir=DEST_DIR] MAP_file..."
    parser = OptionParser(usage=usage,
        description="Converts OziExplorer's .MAP file into GDAL .VRT format")
    parser.add_option("-d", "--debug", action="store_true", dest="debug")
    parser.add_option("-q", "--quiet", action="store_true", dest="quiet")
    parser.add_option("-c", "--cut", action="store_true", 
        help='cut the raster as per Moving Map border line')
    parser.add_option("-t", "--dest-dir", dest="dest_dir", default='',
        help='destination directory (default: current)')
    parser.add_option("--no-data", dest="no_data", default='',
        help='set nodata masking values for input bands, separated by commas')

    (options, args) = parser.parse_args()
    if not args:
        parser.error('No input file(s) specified')
    logging.basicConfig(level=logging.DEBUG if options.debug else 
        (logging.ERROR if options.quiet else logging.INFO))

    for f in args:
        map2vrt(f)

